# server.py

import socket
import time
from typing import Dict, Optional, List, Tuple, Any, Callable, Union
import threading

from socketsio.protocols import BaseProtocol
from socketsio.sockets import Socket

__all__ = [
    "Server"
]

Connection = socket.socket
Address = Tuple[str, int]
Action = Callable[[Connection, Address, BaseProtocol], Any]
Clients = Dict[Tuple[str, int], List[Connection]]

class Server(Socket):
    """A class to represent the server object."""

    DELAY = 0.0001

    def __init__(
            self,
            protocol: BaseProtocol,
            connection: Optional[Connection] = None,
            clients: Optional[Union[Clients, bool]] = None
    ) -> None:
        """
        Defines the attributes of a server.

        :param connection: The socket object of the server.
        :param protocol: The communication protocol object.
        :param clients: The client's collection.
        """

        super().__init__(connection=connection, protocol=protocol)

        if clients in (True, None):
            clients = {}

        elif clients is False:
            clients = None
        # end if

        self.clients: Optional[Clients] = clients

        self._listening = False
        self._bound = False

        self.handling = False

        self._address: Optional[Address] = None
    # end __init__

    @property
    def listening(self) -> bool:
        """
        Returns the value of an active listening process.

        :return: The boolean flag.
        """

        return self._listening
    # end listening

    @property
    def bound(self) -> bool:
        """
        Returns the value of a bounded connection.

        :return: The boolean flag.
        """

        return self._bound
    # end bound

    @property
    def address(self) -> Address:
        """
        Returns the ip and port of the binding.

        :return: The address tuple.
        """

        return self._address
    # end address

    def bind(self, address: Address) -> None:
        """
        Binds the connection of the server.

        :param address: The address to bind to.
        """

        self.connection.bind(address)

        self._address = address

        self._bound = True
    # end bind

    def validate_binding(self) -> None:
        """Validates the binding of the socket."""

        if not self.bound:
            raise ValueError("Cannot start listening before binding.")
        # end if
    # end validate_binding

    def validate_listening(self) -> None:
        """Validates the binding of the socket."""

        if not self.listening:
            self.listen()
        # end if
    # end validate_binding

    def accept(self) -> Tuple[socket.socket, Address]:
        """
        Returns the connection and address from the accepted client.

        :return: The sockets object and the address.
        """

        self.validate_listening()

        return self.connection.accept()
    # end accept

    def handle(
            self,
            protocol: Optional[BaseProtocol] = None,
            action: Optional[Action] = None,
            clients: Optional[Clients] = None,
            remove: Optional[bool] = True
    ) -> None:
        """
        Sends a message to the client by its connection.

        :param protocol: The protocol to use for sockets communication.
        :param action: The action to call.
        :param clients: The client's collection.
        :param remove: The value to remove clients after each response.
        """

        if clients is None:
            clients = self.clients
        # end if

        protocol = protocol or self.protocol
        action = action or self.action

        connection, address = self.accept()

        if clients is not None:
            clients.setdefault(address, []).append(connection)
        # end if

        action(connection, address, protocol)

        if remove and (clients is not None):
            clients.get(address).remove(connection)

            if not clients.get(address):
                clients.pop(address)
            # end if
        # end if
    # end handle

    def action(
            self,
            connection: Connection,
            address: Address,
            protocol: BaseProtocol
    ) -> None:
        """
        Sets or updates clients data in the clients' container .

        :param protocol: The protocol to use for sockets communication.
        :param connection: The socket object of the server.
        :param address: The address of the connection.
        """
    # end action

    def listen(self) -> None:
        """Listens to clients."""

        self.validate_binding()

        self.connection.listen()

        self._listening = True
    # end listen

    def _serve(
            self,
            protocol: Optional[BaseProtocol] = None,
            action: Optional[Action] = None,
            clients: Optional[Clients] = None,
            remove: Optional[bool] = True
    ) -> None:
        """
        Runs the threads to serving_loop to clients with requests.

        :param action: The action to call.
        :param protocol: The protocol to use for sockets communication.
        :param clients: The client's collection.
        :param remove: The value to remove clients after each response.
        """

        if clients is None:
            clients = self.clients
        # end if

        self.validate_listening()

        self.handling = True

        while self.handling:
            threading.Thread(
                target=lambda: self.handle(
                    clients=clients, remove=remove,
                    protocol=protocol, action=action
                )
            ).start()
        # end while
    # end _serve

    def serve(
            self,
            protocol: Optional[BaseProtocol] = None,
            action: Optional[Action] = None,
            clients: Optional[Clients] = None,
            remove: Optional[bool] = True,
            block: Optional[bool] = True
    ) -> None:
        """
        Runs the threads to serving_loop to clients with requests.

        :param action: The action to call.
        :param protocol: The protocol to use for sockets communication.
        :param clients: The client's collection.
        :param remove: The value to remove clients after each response.
        :param block: The value to block the process.
        """

        parameters = dict(
            protocol=protocol, action=action,
            clients=clients, remove=remove
        )

        if block:
            self._serve(**parameters)

        else:
            threading.Thread(
                target=lambda: self._serve(**parameters)
            ).start()

            self.await_handling()
        # end if
    # end serve

    def await_handling(self) -> None:
        """Awaits the start of the handling process."""

        while not self.handling:
            time.sleep(self.DELAY)
        # end while
    # end await_handling
# end Server


# protocols.py

import socket
from typing import Optional, Tuple, List
from abc import ABCMeta, abstractmethod

from represent import represent

Connection = socket.socket
Address = Tuple[str, int]

__all__ = [
    "BufferedProtocol",
    "BaseProtocol",
    "TCP",
    "UDP",
    "BCP",
    "base_socket",
    "tcp_socket",
    "udp_socket",
    "bluetooth_socket"
]

def base_socket() -> Connection:
    """
    Returns a new base socket object.

    :return: The socket object.
    """

    return socket.socket()
# end base_socket

def tcp_socket() -> Connection:
    """
    Returns a new TCP socket object.

    :return: The socket object.
    """

    return socket.socket(socket.AF_INET, socket.SOCK_STREAM)
# end tcp_socket

def udp_socket() -> Connection:
    """
    Returns a new UDP socket object.

    :return: The socket object.
    """

    return socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
# end udp_socket

def bluetooth_socket() -> socket.socket:
    """
    Sends a request through the bluetooth sockets.

    :return: The client object.
    """

    return socket.socket(
        socket.AF_BLUETOOTH, socket.SOCK_STREAM, socket.BTPROTO_RFCOMM
    )
# end bluetooth_socket

@represent
class BaseProtocol(metaclass=ABCMeta):
    """Defines the basic parameters for the communication."""

    @staticmethod
    def socket() -> Connection:
        """
        Returns a new base socket object.

        :return: The socket object.
        """

        return base_socket()
    # end socket

    @abstractmethod
    def send(
            self,
            connection: Connection,
            data: bytes,
            address: Optional[Address] = None
    ) -> None:
        """
        Sends a message to the client or server by its connection.

        :param data: The message to send to the client.
        :param connection: The sockets' connection object.
        :param address: The address of the sender.
        """
    # end send

    @abstractmethod
    def receive(
            self,
            connection: Connection,
            address: Optional[Address] = None,
            buffer: Optional[int] = None
    ) -> bytes:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param address: The address of the sender.
        :param buffer: The buffer size to collect.

        :return: The received message from the server.
        """
    # end receive
# end BaseProtocol

class BufferedProtocol(BaseProtocol, metaclass=ABCMeta):
    """Defines the basic parameters for the communication."""

    SIZE = 1024

    def __init__(self, size: Optional[int] = None) -> None:
        """
        Defines the attributes of the protocol.

        :param size: The buffer size.
        """

        if size is None:
            size = self.SIZE
        # end if

        self.size = size
    # end __init__
# end BufferedProtocol

class TCP(BufferedProtocol):
    """Defines the basic parameters for the communication."""

    def send(
            self,
            connection: Connection,
            data: bytes,
            address: Optional[Address] = None
    ) -> None:
        """
        Sends a message to the client or server by its connection.

        :param data: The message to send to the client.
        :param connection: The sockets' connection object.
        :param address: The address of the sender.
        """

        connection.send(data)
    # end send

    def receive(
            self,
            connection: Connection,
            address: Optional[Address] = None,
            buffer: Optional[int] = None
    ) -> bytes:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param address: The address of the sender.
        :param buffer: The buffer size to collect.

        :return: The received message from the server.
        """

        return connection.recv(buffer or self.size)
    # end receive
# end TCP

class UDP(BufferedProtocol):
    """Defines the basic parameters for the communication."""

    def send(
            self,
            connection: Connection,
            data: bytes,
            address: Optional[Address] = None
    ) -> None:
        """
        Sends a message to the client or server by its connection.

        :param data: The message to send to the client.
        :param connection: The sockets' connection object.
        :param address: The address of the sender.
        """

        if address is None:
            raise ValueError("address must be a tuple of ip and port.")
        # end if

        connection.sendto(data, address)
    # end send

    def receive(
            self,
            connection: Connection,
            address: Optional[Address] = None,
            buffer: Optional[int] = None
    ) -> bytes:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param address: The address of the sender.
        :param buffer: The buffer size to collect.

        :return: The received message from the server.
        """

        if address is None:
            raise ValueError("address must be a tuple of ip and port.")
        # end if

        return connection.recvfrom(buffer or self.size)[0]
    # end receive
# end UDP

class BCP(BaseProtocol):
    """Defines the basic parameters for the communication."""

    HEADER = 32

    def __init__(self, protocol: BufferedProtocol) -> None:
        """
        Defines the base protocol.

        :param protocol: The base protocol object to use.
        """

        self.protocol = protocol
    # end __init__

    def socket(self) -> Connection:
        """
        Returns a new base socket object.

        :return: The socket object.
        """

        return self.protocol.socket()
    # end socket

    def send(
            self,
            connection: Connection,
            data: bytes,
            address: Optional[Address] = None
    ) -> None:
        """
        Sends a message to the client or server by its connection.

        :param data: The message to send to the client.
        :param connection: The sockets' connection object.
        :param address: The address of the sender.
        """

        message_len = len(data)

        size = self.protocol.size

        length_message = (
            str(message_len) +
            (" " * (self.HEADER - len(str(message_len))))
        ).encode()

        self.protocol.send(
            connection=connection,
            data=length_message,
            address=address
        )

        if size >= message_len:
            return self.protocol.send(
                connection=connection,
                data=data,
                address=address
            )
        # end if

        total = 0

        for i in range(0, message_len, size):
            self.protocol.send(
                connection=connection,
                data=data[i:i + size],
                address=address
            )

            total += len(data[i:i + size])

            if total == message_len:
                return
            # end if
        # end for

        if message_len % size:
            self.protocol.send(
                connection=connection,
                data=data[-(message_len % size):],
                address=address
            )
        # end if
    # end send

    def receive(
            self,
            connection: Connection,
            address: Optional[Address] = None,
            buffer: Optional[int] = None
    ) -> bytes:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param address: The address of the sender.
        :param buffer: The buffer size to collect.

        :return: The received message from the server.
        """

        length_message = self.protocol.receive(
            connection=connection,
            address=address,
            buffer=self.HEADER
        ).decode()

        if not length_message or length_message == '0':
            return b''
        # end if

        message_len = int(length_message[:length_message.find(" ")])

        size = buffer or self.protocol.size

        if size >= message_len:
            return self.protocol.receive(
                connection=connection,
                address=address,
                buffer=message_len
            )
        # end if

        data: List[bytes] = []

        for _ in range(message_len // size):
            data.append(
                self.protocol.receive(
                    connection=connection,
                    address=address,
                    buffer=size
                )
            )
        # end for

        if message_len % size:
            data.append(
                self.protocol.receive(
                    connection=connection,
                    address=address,
                    buffer=message_len % size
                )
            )
        # end if

        return b''.join(data)
    # end receive
# end BCP